@@ -196,13 +196,11 @@ impl<T: ZByteReaderTrait> JpegDecoder<T> {
                 &self.z_order[..usize::from(self.num_scans)]
             );
 
+            trace!("Decoding MCU width: {mcu_width}, height: {mcu_height}");
+
             for i in 0..mcu_height {
-                // Report if we have no more bytes
-                // This may generate false negatives since we over-read bytes
-                // hence that why 37 is chosen(we assume if we over-read more than 37 bytes, we have a problem)
-                if stream.overread_by > 37
-                // favourite number :)
-                {
+                if stream.overread_by > 0 {
+                    pixels.get_mut(pixels_written..).map(|v| v.fill(128));
                     if self.options.strict_mode() {
                         return Err(DecodeErrors::FormatStatic("Premature end of buffer"));
                     };
@@ -271,6 +269,7 @@ impl<T: ZByteReaderTrait> JpegDecoder<T> {
                     McuContinuation::AnotherSos => continue 'sos,
                     McuContinuation::Terminate => {
                         warn!("Got terminate signal, will not process further");
+                        pixels.get_mut(pixels_written..).map(|v| v.fill(128));
                         return Ok(());
                     }
                 }
@@ -388,12 +387,64 @@ impl<T: ZByteReaderTrait> JpegDecoder<T> {
     fn decode_mcu_width<const PROGRESSIVE: bool>(
         &mut self, mcu_width: usize, mcu_height: usize, tmp: &mut [i32; 64],
         stream: &mut BitStream, progressive: &mut [Vec<i16>; 4]
+    ) -> Result<McuContinuation, DecodeErrors> {
+        let is_one_by_one = !self.scan_subsampled;
+
+        // The definition of MCU depends on the sampling factor of involved scans. When components
+        // have different factors then each Minimal-Coding-Unit is the least common multiple such
+        // that we have an integer number of blocks from each component. But the decoding of these
+        // components differs from it otherwise, we need an inner loop with a dynamic amount of
+        // coefficients per component, whereas otherwise we have exactly one block of coefficients
+        // encoded for each component in the bitstream order.
+        //
+        // We statically specialize on this to improve code generation of the common case a little
+        // bit. We could also special case common sub-sampling cases but be mindful of code bloat.
+        if is_one_by_one {
+            self.inner_decode_mcu_width::<PROGRESSIVE, false>(
+                mcu_width,
+                mcu_height,
+                tmp,
+                stream,
+                progressive
+            )
+        } else {
+            self.inner_decode_mcu_width::<PROGRESSIVE, true>(
+                mcu_width,
+                mcu_height,
+                tmp,
+                stream,
+                progressive
+            )
+        }
+    }
+
+    // Inline-never ensures we do get this function optimize on its own, into two different
+    // versions, without the optimizer tripping up over the complexity that comes with the
+    // constant folding. And constant folding is quite important for performance here as
+    // when `not SAMPLED` then the inner loop has exactly one iteration per component in
+    // the scan. The difference was ~1% or a bit more.
+    fn inner_decode_mcu_width<const PROGRESSIVE: bool, const SAMPLED: bool>(
+        &mut self, mcu_width: usize, mcu_height: usize, tmp: &mut [i32; 64],
+        stream: &mut BitStream, progressive: &mut [Vec<i16>; 4]
     ) -> Result<McuContinuation, DecodeErrors> {
         let z_order = self.z_order;
+        let z_scans = &z_order[..usize::from(self.num_scans)];
+
+        // How much of the head of `tmp` was written by the last MCU decoding? We only check for
+        // two different cases and not all possible outcomes as this is only used to optimize the
+        // bytes written in `fill`. Since the clobber happens in UNZIGZAG order we'd be straddling
+        // most cache lines anyways even if we did a partial write with the exact length of the
+        // coefficient data which was written into `tmp`.
+        let mut clobber_more_than_4x4 = true;
 
         for j in 0..mcu_width {
             // iterate over components
-            for &k in &z_order[..usize::from(self.num_scans)] {
+            for &k in z_scans {
+                // we made this loop body massive due to several different paths that depend on
+                // static conditions. Note we (potentially) call into other functions so the
+                // compiler will not unroll anything here anyways. The gains from separating
+                // differently optimized loop bodies are much greater than a single additional jump
+                // here.
                 let component = &mut self.components[k];
 
                 let dc_table = self.dc_huffman_tables[component.dc_huff_table % MAX_COMPONENTS]
@@ -413,23 +464,46 @@ impl<T: ZByteReaderTrait> JpegDecoder<T> {
                     &mut component.raw_coeff
                 };
 
+                let component_samples_needed = component.needed;
+
                 // If image is interleaved iterate over scan components,
                 // otherwise if it-s non-interleaved, these routines iterate in
                 // trivial scanline order(Y,Cb,Cr)
-                for v_samp in 0..component.vertical_sample {
-                    for h_samp in 0..component.horizontal_sample {
-                        // Fill the array with zeroes, decode_mcu_block expects
-                        // a zero based array.
-                        tmp.fill(0);
+                //
+                // Turn the bounds into a compile time constant for a common special case. This
+                // allows the compiler to unroll the loop and then do a bunch of interleaving.
+                let v_step = if SAMPLED { 0..component.vertical_sample } else { 0..1 };
+
+                for v_samp in v_step {
+                    let h_step = if SAMPLED { 0..component.horizontal_sample } else { 0..1 };
+
+                    for h_samp in h_step {
+                        let result = if component_samples_needed {
+                            // Fill the array with zeroes, decode_mcu_block expects
+                            // a zero based array. Clobber is in zig-zag order though.
+                            // Writing consecutive entries is basically free in terms
+                            // of memory throughput so we opt for a larger power of
+                            // two which lets the compiler turn this into a repeated
+                            // write of a zeroed vector register, which does not have
+                            // any branches, instead of a more difficult pattern where
+                            // we attempt to overwrite exactly one coefficient.
+                            let clobber_len = if !clobber_more_than_4x4 { 32 } else { 64 };
+
+                            tmp[..clobber_len].fill(0);
+
+                            stream.decode_mcu_block(
+                                &mut self.stream,
+                                dc_table,
+                                ac_table,
+                                qt_table,
+                                tmp,
+                                &mut component.dc_pred
+                            )
+                        } else {
+                            // We do not touch tmp so there is no need to reset it.
+                            stream.discard_mcu_block(&mut self.stream, dc_table, ac_table)
+                        };
 
-                        let result = stream.decode_mcu_block(
-                            &mut self.stream,
-                            dc_table,
-                            ac_table,
-                            qt_table,
-                            tmp,
-                            &mut component.dc_pred
-                        );
                         // If an error occurs we can either propagate it
                         // as an error or print it and call terminate.
                         //
@@ -437,16 +511,22 @@ impl<T: ZByteReaderTrait> JpegDecoder<T> {
                         // even if its bad, matching browsers.
                         //
                         // See example in https://github.com/etemesi254/zune-image/issues/293
-                        if result.is_err() {
+                        let len = if let Ok(len) = result {
+                            len
+                        } else {
+                            // result.is_err()
                             return if self.options.strict_mode() {
                                 Err(result.err().unwrap())
                             } else {
                                 error!("{}", result.err().unwrap());
                                 Ok(McuContinuation::Terminate)
                             };
-                        }
+                        };
+
+                        if component_samples_needed {
+                            // tmp was only written partially, note that len is in ZigZag order.
+                            clobber_more_than_4x4 = len > 10;
 
-                        if component.needed {
                             let idct_position = {
                                 // derived from stb and rewritten for my tastes
                                 let c2 = v_samp * 8;
@@ -456,8 +536,15 @@ impl<T: ZByteReaderTrait> JpegDecoder<T> {
                             };
 
                             let idct_pos = channel.get_mut(idct_position..).unwrap();
-                            //  call idct.
-                            (self.idct_func)(tmp, idct_pos, component.width_stride);
+
+                            if len <= 1 {
+                                (self.idct_1x1_func)(tmp, idct_pos, component.width_stride);
+                            } else if len <= 10 {
+                                (self.idct_4x4_func)(tmp, idct_pos, component.width_stride);
+                            } else {
+                                //  call idct.
+                                (self.idct_func)(tmp, idct_pos, component.width_stride);
+                            }
                         }
                     }
                 }
@@ -475,6 +562,12 @@ impl<T: ZByteReaderTrait> JpegDecoder<T> {
             }
         }
 
+        self.check_stream_marker_after_mcu_width(stream)
+    }
+
+    fn check_stream_marker_after_mcu_width(
+        &mut self, stream: &mut BitStream
+    ) -> Result<McuContinuation, DecodeErrors> {
         // After all interleaved components, that's an MCU
         // handle stream markers
         //
@@ -528,6 +621,7 @@ impl<T: ZByteReaderTrait> JpegDecoder<T> {
 
         Ok(McuContinuation::Ok)
     }
+
     // handle RST markers.
     // No-op if not using restarts
     // this routine is shared with mcu_prog
@@ -593,7 +687,11 @@ impl<T: ZByteReaderTrait> JpegDecoder<T> {
                     // iterate over each line, since color-convert needs only
                     // one line
                     for (j, samp) in raw_samples.iter_mut().enumerate().take(comp_len) {
-                        *samp = &samples[j][pos * padded_width..(pos + 1) * padded_width];
+                        let temp = &samples[j].get(pos * padded_width..(pos + 1) * padded_width);
+                        if temp.is_none() {
+                            return Err(DecodeErrors::FormatStatic("Missing samples"));
+                        }
+                        *samp = temp.unwrap();
                     }
                     color_convert(
                         &raw_samples,
